//
//  Cluster.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#include "Cluster.h"

double Cluster::total_count = 0.0;
double Cluster::total_area = 0.0;

Cluster::
Cluster( Region* in_region, 
        Distance* in_dist )
{
    core.clear();
    core.insert( in_region );
    d_convex_solid.insert( in_region );
    count = in_region->Count();
    area = in_region->Area();
    peak = in_region;
    dist = in_dist;
    CalcAreaShare();
}    


Cluster& Cluster::
operator=( Cluster& in_cluster )
{
    SetPeak( in_cluster.Peak() );
    count = in_cluster.Count();
    total_count = in_cluster.TotalCount();
    area = in_cluster.Area();
    //  total_area = in_cluster.TotalArea();
    CalcAreaShare();
    dist = in_cluster.Dist();
    
    d_convex_solid.clear();
    
    for ( set<Region*>::iterator i1 = 
         in_cluster.DConvexSolid().begin();
         i1 != in_cluster.DConvexSolid().end(); ++i1 )
    {
        d_convex_solid.insert( *i1 );
    }
    
    core.clear();
    for ( set<Region*>::iterator i1 = in_cluster.Core().begin();
         i1 != in_cluster.Core().end(); ++i1 )
    {
        core.insert( *i1 );
    }
    
    return *this;
}

bool Cluster::
operator==( Cluster& in_cluster ) 
{
    if ( ( core.size() != in_cluster.Core().size() ) ||
        ( d_convex_solid.size() !=
         in_cluster.DConvexSolid().size() ) )
        return false;
    
    bool result = true;
    
    for ( set<Region*>::iterator i1 = core.begin();
         i1 != core.end(); ++i1 )
    {
        if ( in_cluster.Core().find( *i1 ) == 
            in_cluster.Core().end() )
        {
            result = false;
            break; 
        }
    }
    return result;
}

bool Cluster::
operator!=( Cluster& in_cluster )
{
    return !( operator==( in_cluster ) );
}

void Cluster::
CalcCountAndArea()
{
    double val_count = 0.0;
    double val_area = 0.0;
    
    for ( set<Region*>::iterator i1 = d_convex_solid.begin();
         i1 != d_convex_solid.end(); ++i1 )
    {
        val_count += (**i1).Count();
        val_area += (**i1).Area();
    }
    
    count = val_count;
    area = val_area;
}


bool Cluster::
IfCoreMember( Region* in_region ) const
{
    if ( core.find( in_region ) != core.end() )
        return true;
    else
        return false;
}

bool Cluster::
IfDConvexSolidMember( Region* in_region ) const
{
    if ( d_convex_solid.find( in_region ) !=  d_convex_solid.end() )
        return true;
    else
        return false;
}

void Cluster::
AddDConvexSolid( Region* in_region )
{
    d_convex_solid.insert( in_region );
}

void Cluster::
AddCore( Region* in_region )
{
    core.insert( in_region );
}

double Cluster::
DistanceTo( const Region* target ) const
{
    double min_val = (*dist)( Location(), target->Location() );
    
    for ( set<Region*>::iterator i1 =
         core.begin(); i1 != core.end(); ++i1 )
    {
        double tmp = (*dist)( (**i1).Location(), target->Location() );
        
        if ( min_val > tmp )
            min_val = tmp;
    }
    
    return min_val;
}


double Cluster::
DistanceToFringe( const Region* target ) const
{
    double min_val = (*dist)( Location(), target->Location() );
    
    for ( set<Region*>::iterator i1 =
         d_convex_solid.begin(); i1 != d_convex_solid.end(); ++i1 )
    {
        double tmp = (*dist)( (**i1).Location(), target->Location() );
        
        if ( min_val > tmp )
            min_val = tmp;
    }
    
    return min_val;
}


double Cluster::
DistanceTo( const Cluster* target ) const
{
    
    double val = (*dist)( Location(), target->Location() );
    
    for ( set<Region*>::iterator i1 = core.begin(); 
         i1 != core.end(); ++i1 )
    {
        double tmp = target->DistanceTo( *i1 );
        
        if ( val > tmp )
            val = tmp;
    }
    
    return val;
}

double Cluster::
DistanceToFringe( const Cluster* target ) const
{
    
    double val = (*dist)( Location(), target->Location() );
    
    for ( set<Region*>::iterator i1 = d_convex_solid.begin(); 
         i1 != d_convex_solid.end(); ++i1 )
    {
        double tmp = target->DistanceToFringe( *i1 );
        
        if ( val > tmp )
            val = tmp;
    }
    
    return val;
}


void Cluster::
FindNeighborRegions( set<Region*>& potential_regions,
                    list<Region*>& neighbor_regions,
                    double max_dist )
{
    
    list< pair< Region*, double> > ordered_regions;
    
    for ( set<Region*>::iterator i1 = potential_regions.begin();
         i1 != potential_regions.end(); ++i1 )
    {
        if ( (**i1).Count() < 0.5 )
            continue;
        
        if ( d_convex_solid.find( *i1 ) != d_convex_solid.end() )
            continue;
        
        double val = DistanceTo(*i1);
        
        if (  val > max_dist )
            continue;
        // If outside the search range.
        
        ordered_regions.push_back( make_pair(*i1,val) );
    }
    
    ordered_regions.sort( ValueCompareRegion() );
    
    for ( list< pair< Region*, double > >::
         iterator i1 = ordered_regions.begin(); 
         i1 != ordered_regions.end(); ++i1 )
    {
        neighbor_regions.push_back( i1->first );
    }
}

void Cluster::
CalcSpacing( set<Cluster*>& in_clusters )
{
    spacing.clear();
    
    for ( set<Cluster*>::iterator i1 = in_clusters.begin();
         i1 != in_clusters.end(); ++i1 )
    {
        if ( (**i1).Location() == Location() )
            continue;
        
        spacing[*i1] = DistanceTo( *i1 );
    }
}

void Cluster::
FindNeighborClusters( set<Cluster*>& all_clusters )
{
    list<Cluster*> ordered_clusters;
    neighbors.clear();
    
    for ( set<Cluster*>::iterator i = all_clusters.begin();
         i != all_clusters.end(); ++i )
    {
        if ( (**i).Location() == Location() )
            continue;
        
        ordered_clusters.push_back( *i );
    }
    
    ClusterCompareProximity proximity( this );
    
    ordered_clusters.sort( proximity );
    neighbors.insert( ordered_clusters.front() );
    ordered_clusters.pop_front();
    
    for ( list<Cluster*>::iterator i1 = ordered_clusters.begin();
         i1 != ordered_clusters.end(); ++i1 )
    {
        bool if_neighbor = true;
        
        for ( set<Cluster*>::iterator i2 = neighbors.begin();
             i2 != neighbors.end(); ++i2 )
        {
            if ( spacing[*i1] > (**i1).Spacing()[*i2] )
            {
                if_neighbor = false;
                break;
            }
        }
        if ( if_neighbor )
            neighbors.insert( *i1 );
    }
}


void Cluster::
DConvexify()
{
    set<Region*> path;
    // All the basic regions to be added by D-convexification
    // will be contained in "path".
    
    for ( set<Region*>::iterator i1 = core.begin(); 
         i1 != core.end(); ++i1 )
	{
        
        // Identify the basic regions at the edge of the cluster.
        bool if_inside = true;
        
        for ( set<Region*>::iterator i2 = (**i1).Neighbors().begin();
             i2 != (**i1).Neighbors().end(); ++i2 )
	    {
            if ( d_convex_solid.find( *i2 ) == 
                d_convex_solid.end() )
            {
                if_inside = false;
                break;
            }
	    }
        
        if ( if_inside )
            continue;
        
        // Include all the basic regions on the shortest path 
        // from the edge region to all else in the cluster.
        
        for ( set<Region*>::iterator i2 = i1;
             i2 != core.end(); ++i2 )
	    {
            (**i1).FindShortestPath( path, *i2 );
	    }
	}
    
    for ( set<Region*>::iterator i1 = path.begin();
         i1 != path.end(); ++i1 )
	{
        if ( (**i1).Count() > 0.5 )
            core.insert( *i1 );
        
        d_convex_solid.insert( *i1 );
	}
}

void Cluster::
DConvexifyFull()
{
    unsigned long dcs_size;
    
    do 
    {
        dcs_size = d_convex_solid.size();
        
        set<Region*> path;
        // All the basic regions to be added by D-convexification
        // will be contained in "path".
         
        for ( set<Region*>::iterator i1 = d_convex_solid.begin(); 
             i1 != d_convex_solid.end(); ++i1 )
        {
            bool if_inside = true;
            
            for ( set<Region*>::iterator i2 = (**i1).Neighbors().begin();
                 i2 != (**i1).Neighbors().end(); ++i2 )
            {
                if ( d_convex_solid.find( *i2 ) == 
                    d_convex_solid.end() )
                {
                    if_inside = false;
                    break;
                }
            }
            
            if ( if_inside )
                continue;
            
            // Include all the basic regions on the shortest path 
            // from the edge region to all else in the cluster.
            
            for ( set<Region*>::iterator i2 = i1;
                 i2 != d_convex_solid.end(); ++i2 )
            {
                (**i1).FindShortestPath( path, *i2 );
            }
        }
        
        for ( set<Region*>::iterator i1 = path.begin();
             i1 != path.end(); ++i1 )
        {
            if ( (**i1).Count() > 0.5 )
                core.insert( *i1 );
            
            d_convex_solid.insert( *i1 );
        }
    }
    while ( d_convex_solid.size() > dcs_size );
}


void Cluster::
Solidify( set<Region*>& all_regions )
{
    for ( set<Region*>::iterator i1 = all_regions.begin();
         i1 != all_regions.end(); ++i1 )
    {
        if ( d_convex_solid.find( *i1 ) !=
            d_convex_solid.end() )
            continue;
        
        set<Region*> path;
        
        bool if_outside =(**i1).IfOutSideConvexHull( d_convex_solid, path );
        // Find if region i1 is in a hole within the d-convex set.
        
        if ( !if_outside )
        {
            if ( (**i1).Count() > 0.5 )
                core.insert( *i1 );
            
            d_convex_solid.insert( *i1 );
        }
    }
}

void Cluster::
DConvexSolidify( set<Region*>& all_regions )
{
    unsigned long core_size;
    
    do 
    {
        core_size = core.size();
        DConvexify();
        Solidify( all_regions );
    }
    while( core.size() > core_size );
}

void Cluster::
DConvexSolidifyFull( set<Region*>& all_regions )
{
    unsigned long dcs_size;
    
    do 
    {
        dcs_size = d_convex_solid.size();
        DConvexifyFull();
        Solidify( all_regions );
    }
    while( d_convex_solid.size() > dcs_size );
}


void Cluster::
FindConvergentClosure( set<Region*>& all_regions, 
                      set<Cluster*>& all_clusters,
                      set<Cluster*>& absorbed_clusters )
{
    unsigned long default_size;
    
    do
    {
        default_size = d_convex_solid.size();
        DConvexSolidify( all_regions );
        
        for ( set<Cluster*>::iterator i1 = all_clusters.begin();
             i1 != all_clusters.end(); ++i1 )
        {
            if ( (**i1).Location() == Location() || 
                absorbed_clusters.find( *i1 ) != absorbed_clusters.end() )
                continue;
            
            if ( IfSeparated( **i1 ) )
                continue;
            else
            {
                absorbed_clusters.insert( *i1 );
                Expand( *i1 );
            }
        }
    }
    while ( default_size < d_convex_solid.size() );
}



void Cluster::
Expand( Region* target )
{
    if ( target->Count() < 0.5 )
        return;
    
    core.insert( target );
    d_convex_solid.insert( target );
}

void Cluster::
Expand( Cluster* target )
{
    for ( set<Region*>::iterator i1 = target->Core().begin();
         i1 != target->Core().end(); ++i1 )
    {
        core.insert( *i1 );
    }
    
    for ( set<Region*>::iterator i1 = target->DConvexSolid().begin();
         i1 != target->DConvexSolid().end(); ++i1 )
    {
        d_convex_solid.insert( *i1 );
    }
}



bool Cluster::
IfSeparated( Cluster& in_cluster )
{
    for ( set<Region*>::iterator i1 = 
         d_convex_solid.begin(); 
         i1 != d_convex_solid.end(); ++i1 )
    {
        if ( in_cluster.DConvexSolid().find( *i1 ) != 
            in_cluster.DConvexSolid().end() ) 
            return false;
    }
    return true;
}

bool Cluster::
IfAdjacent( Region* in_region ) const
{
    for ( set<Region*>::iterator i1 = d_convex_solid.begin();
         i1 != d_convex_solid.end(); ++i1 )
    {
        if ( (**i1).Neighbors().find( in_region ) != 
            (**i1).Neighbors().end() )
            return true;
    }
    
    return false;
}

bool ClusterCompareBIC::
operator()( const Cluster* c1, const Cluster* c2 ) const
{
    return ( c1->BIC() > c2->BIC() );
}


bool ClusterCompareLocation::
operator()( const Cluster* c1, const Cluster* c2 ) const
{
    return ( c1->Location() < c2->Location() );
}

bool ClusterCompareProximity::
operator()( const Cluster* c1, const Cluster* c2 ) const
{
    return ( reference_cluster->DistanceTo(c1) < 
            reference_cluster->DistanceTo(c2) );
} 


