//
//  Cluster.h
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_Cluster_h
#define DetectClusters_Cluster_h

#include <utility>
#include <set>
#include <list>
#include <map>
#include <vector>
#include <algorithm>
#include <cmath>
#include <iostream>
#include "Region.h"
#include "Distance.h"
#include "StringManipulator.h"

using namespace std;

class Cluster
{
private:
    Region* peak; 
    // Cluster ID (First-identified basic region)
    
    double area;
    // Economic area size
    
    double area_share;
    // = area / total_area
    
    double count;
    // Establishment counts in this cluster
    
    set<Region*> core;
    // Set of basic regions with establishments in this cluster
    
    set<Region*> d_convex_solid;
    // Set of basic regions in this cluster
    // = convex solidified set of core regions
    
    double bic;
    // BIC contribution of this cluster within the cluster scheme
    
    map<Cluster*,double> spacing;
    // Distance to other clusters.
    
    set<Cluster*> neighbors;
    // Set of neighbor clusters.
    // If cluster r is a neighbor of this cluster, then it is either the closest neighor, 
    // or the distance between cluster r and other neighbor cluster s is larger than
    // the distance between cluster s and this cluster.
    
    Distance* dist;
    // Inter-regional distance data
    // dist(x,y) = distance between regions x and y
    //             (DBL_MAX if x and y are disconnected).
    
public:
    static double total_area;
    // national economic area
    
    static double total_count;
    // total establishment counts in the industry
    
    
    int Location() const { return peak->Location(); }
    Region* Peak() const { return peak; }
    void SetPeak( Region* in_peak ) { peak = in_peak; }

    
    const set<Region*>& Core() const { return core; }
    set<Region*>& DConvexSolid() { return d_convex_solid; }

    void AddCore( Region* in_region );
    // Expand core by adding in_region.
    
    void AddDConvexSolid( Region* in_region );
    // Expand d_convex_solid by adding in_region.
    
    Distance* Dist() const { return dist; }
    

    double Area() const { return area; }
    double TotalArea() const { return total_area; }
    //    void CalcTotalArea();
    double AreaShare() const { return area_share; }
    void CalcAreaShare() { area_share = area / total_area; }
    
    double Count() const { return count; }
    double TotalCount() const { return total_count; }

    void CalcCountAndArea();
    // Update count and area
    
    map<Cluster*,double>& Spacing() { return spacing; }
    void CalcSpacing( set<Cluster*>& in_clusters );
    // Compute distances to clusters in in_clusters.
  
    double BIC() const { return bic; }
    void BIC( double in_val ) { bic = in_val; }
    /// BIC contribution in the cluster system
    
    bool IfCoreMember( Region* in_region ) const;
    // If core includes in_region
    
    bool IfDConvexSolidMember( Region* in_region ) const;
    // If D-convex solid includes in_region
    
    bool IfSeparated( Cluster& in_cluster );
    // If this cluster is disjoint with in_cluster
    
    bool IfAdjacent( Region* in_region ) const;
    // If this cluster is adjacent to in_region
    
    double DistanceTo( const Region* target ) const;
    // Shortest distance between core regions in this cluster to target region.
    
    double DistanceTo( const Cluster* target ) const;
    // Shortest distance between core regions in target cluster and this cluster.
    
    double DistanceToFringe( const Region* target ) const;
    // Shortest distance between target and this cluster.
    // Note: Cluster = core and filled (under convex-solidification) basic regions.
    
    double DistanceToFringe( const Cluster* target ) const;
    // Shortest distance between target and this cluster.
    // Note: Cluster = core and filled (under convex-solidification) basic regions.
    
    void DConvexSolidfy( set<Region*>& all_regions );
    // If d-convex solidification of the current members overlaps with 
    // other clusters, these are merged to this cluster.
    /// D-convex solidification continues until there is no overlap with 
    // other clusters.
    
    void DConvexify(); 
    // DConvexify wrt core (i.e., regions with establishments)
    // For cluster detection, this approximate version generates a fairly reasonable convex cluters.
    // (Exact d-convexification is given by DConvexifyFull().
    
    void DConvexifyFull(); 
    // The precise version of d-convexification, i.e., d-convexification wrt d_convex_solid.
    // For the d-convexification of a large area, the approximation of d-convexification by DConvexify() is not good.
    // This precise version is used to derive essential containments.
        
    void Solidify( set<Region*>& all_regions );
    // Fill the hole in the cluster (d_convex_solid).
    
    void DConvexSolidify( set<Region*>& all_regions );
    // Version with DConvexify() ( used to detect clusters)
    
    void DConvexSolidifyFull( set<Region*>& all_regions );
    // Version with DConvexifyFull() ( used to find essential containment)
    
    
    void FindConvergentClosure( set<Region*>& all_regions, 
                               set<Cluster*>& all_clusters,
                               set<Cluster*>& absorbed_clusters );
    // Find the convergent d-convex solidification for this cluster,
    // i.e., d-convex solidify all the overlapping clusters.
    // All the absorbed clusters are stored in absorbed_clusters (to be deleted).
    // all_regions : set of all basic regions to be given by ClusterSystem.
    // all_clusters: set of all clusters to be given by ClusterSystem.
    
    void FindNeighborRegions( set<Region*>& potential_regions,
                             list<Region*>& neighbor_regions,
                             double max_dist );
    // Find non-cluster regions with positive counts within
    // max_dist range from this cluster, and store in neighobr_regions
    // in the order of proximity to this cluster.
    // potential_regions = all residual regions outside any clusters
    // (to be given by ClusterSystem).
    
    
    void FindNeighborClusters( set<Cluster*>& all_clusters );
    // Find neighbor clusters (neighbors).
    // (See "neighbors" above for the definition of "neighbor".)

    set<Cluster*>& Neighbors() { return neighbors; }
    
    void Expand( Region* target );
    // Add target region to the cluster (d_convex_solid).
    // (Not D-convex solidfy)
    
    void Expand( Cluster* target );
    // Merge target cluster to this cluster.
    // (Not D-convex solidify)
    

    ///// Operators /////
    
    Cluster& operator=( Cluster& in_cluster);
    // Copies :
    // core and d_convex_solid
    // peak region
    // area and area share
    // count and total count
    
    bool operator==( Cluster& in_cluster );
    // Returns true if core is the same.
    
    bool operator!=( Cluster& in_cluster );
    // !operator==()
    
    ///// Constructors /////
    
    Cluster( Region* in_region, 
            Distance* in_dist = NULL
            );
    // Construct a cluster with a single basic region (in_region).
    
    Cluster( Cluster& in_cluster ) 
    { operator=( in_cluster ); }
    
    ~Cluster() {}
    
};

/// Binary predicates for sorting clusters ///

class ClusterCompareBIC
{
// Compares BIC contributions of clusters within cluster scheme.
public:
    ClusterCompareBIC() {}
    ~ClusterCompareBIC() {}
    bool operator()( const Cluster* c1, const Cluster* c2 ) const;
};

class ClusterCompareLocation
{
// Compares cluster location IDs.
public:
    ClusterCompareLocation() {}
    ~ClusterCompareLocation() {}
    bool operator()( const Cluster* c1, const Cluster* c2 ) const;
};


class ClusterCompareProximity
{
// Compares clusters in terms of the distance to a given cluster (in_cluster).
public:
    Cluster* reference_cluster;
    ClusterCompareProximity( Cluster* in_cluster ) 
    { reference_cluster = in_cluster; }
    
    ~ClusterCompareProximity() {}
    bool operator()( const Cluster* c1, const Cluster* c2 ) const;
};



#endif
