//
//  ClusterBuilder.h
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_ClusterBuilder_h
#define DetectClusters_ClusterBuilder_h


#include <set>
#include <list>
#include <map>
#include <algorithm>
#include <limits>
#include <cmath>
#include <iostream>
#include "Region.h"
#include "Cluster.h"
#include "ClusterSystem.h"
#include "Distance.h"
#include "StringManipulator.h"


using namespace std;

class ClusterBuilder
{
private:
    string id; //Industry ID
    
    ClusterSystem* potential_system;
    ClusterSystem* candidate_new_cluster;
    // Temporary cluster scheme in the search for the best scheme
    
    ClusterSystem* candidate_expansion;
    ClusterSystem* current_system;
    
//    int criterion_choice; 
//    map<int, vector<double>* >* penalty_table; 
    
    set<Region*>* all_regions;
    // all basic regions
    
    Distance* distance;
    // Inter-region distance data
    
    double ref_tc;
    
    double criterion_value;
    
    double max_vacant_radius;
    // maximum range for cluster expansion
    
    double expand_search_step;
    
    int log_count;
    // Count of cluster scheme modification
    
    string eventlog;
    // event log file name
    
    ofstream log_os;
    // Computation log output
    
public:
    
//    enum criterion 
//    { _aic,
//        _bic,
////        _nml,
//        _criterion_end };
//    
    enum event
    { _formation,
        _expansion,
//        _deletion,
        _event_end };
    
    enum logtype
    { _event,
        _summary,
        _main,
        _map,
        _logtype_end };
    
    bool iftakelog[_logtype_end];
    
    static const string summary;
    // summary log file name head
    
    static const string map_data;
    // map-data file name head
    
    static const string core_data;
    // core file name head
    
    static const string event_log;
    // event log file name head
    
    static const string comp_log;
    // computation log file
    
    string out_dir;
    // output directory name
    
//    static const string criterionName[_criterion_end];
    
    static const string eventName[_event_end];
    
    
    bool if_random;
    // Set true for generating spurious cluster schemes.
    
    int sample_id;
    // Id for spurious cluster scheme sample
    
private:
    
    double total_area;
    // National economic area
    
    double total_count;
    // Industry-total establishment count
    
public:
    void SetSampleId( int val ) { sample_id = val; }
    int SampleId() const { return sample_id; }
    
    void SetRandom() { if_random = true; }

    bool IfRandom() const { return if_random; }
    
    const string EventLog() { return eventlog; }
    
    bool IfTakeLog( int log_type ) const { return iftakelog[log_type]; }
    
    void SetLogType( int log_type, bool flagvalue ) 
    { iftakelog[log_type] = flagvalue; }
    
    double Evaluate( ClusterSystem* in_system );
    // Computes BIC for a given cluster scheme (in_system).
    
    Cluster* GenerateFirstCluster();
    // Constructs a single-region cluster scheme with the largest model likelihood (i.e., BIC).

    Cluster* GenerateFirstCluster(Region* initial_region);
    // Constructs a single-region cluster scheme with a given basic region (initial_region).
    
    void GenerateNewCluster( double& new_value,
                            int& new_cluster ); 
    // Find the best cluster formation.
    // The resulting cluster scheme is stored in candidate_new_cluster.
    // new_cluster : the best cluster to form.
    // new_value : BIC value after adding the new_cluster.
    
    Cluster* GenerateNewCluster( Region* target,
                                double& new_value,
                                int& new_cluster );
    // Generate a new single-region cluster at a given region (target),
    // and returns a pointer to this new cluster.
    // new_cluster : location of the new cluster
    // new_value : BIC value after adding the new cluster
    
    
//    /// Functions specific to backward search ///
//    
//    ///// Functions to check potential deletion of a cluster /////
//    
//    double Delete( Cluster* target );
//    // Construct a hypothetical cluster scheme which is identical to
//    // the present scheme except that target cluster is deleted.
//    
//    int Delete( double& new_value,
//               int& deleted_location );
//    // Find the best deletion of a single cluster.
//    
    double ExpandCluster( double& new_value,
                         int& expanded_location,
                         int& consolidated_location );
    
    // Find the best potential expansion of an existing cluster.
    // Returns the BIC value of the best potential expansion.
    // new_value : BIC value of the best potential expansion
    // expanded_location : the location of the best cluster to be expanded
    // consolidated_location : the location of a residual region / cluster 
    // toward which the cluster (expanded_location) is expanded.
    
    
    void GrowSystem( Region* initial_region = NULL);
    // Detects cluster scheme by forward search.
    // It starts with the best single-region cluster (in terms of BIC),
    // then continues growing the scheme either by expanding an existing cluster or 
    // by constructing a single-region cluster, until no further improvement in BIC
    // is possible.
    
    void SetIndustryId( string in_id ) { id = in_id; }
    // Sets industry ID.
    
    void SetOS( const string& out_dir_name );
    // Sets output directory.
    
    void SetEventLog();
    // Creates an event log file.
    // Each event (new formation and expansion of a cluster) is recorded 
    // with corresponding chagnes in likelihood and BIC values.
    
    void SetCompLog();
    // Creates an cluster scheme identification log file.
    // More detailed log (than event log above) will be recorded.
    
//    void SetCriterion( int in_choice ) { criterion_choice = in_choice; }
    
    void WriteFormationLog(// int in_event, 
                       int event_cluster );
    // Writes event log for the new cluster formation.
    // event_cluster : the location of new cluster
    
    void WriteExpansionLog(// int in_event, 
                        int event_cluster,
                        int in_included );
    // Writes event log for cluster expansion.
    // event_cluster : the cluster expanded
    // in_included : consolidated region / cluster
    
    void WriteLog( bool if_final = false );
    // Writes summary log, and cluster compositions (core and d-convex solid).
    // Intermediate outputs are associated with log ID.
    // For random samples (generating spurious cluster schemes), outputs are index by log and sample IDs.
    
    ClusterBuilder( //int in_criterion, 
                   set<Region*>* in_all_regions,
                   double in_expand_step = 0.0,
                   double in_national_area = 0.0,
                   double in_industry_count = 0.0,
                   Distance* in_dist = NULL);
    // in_all_regions : set of all the basic regions.
    // in_expansion_step : search distance-range for cluster expansion
    // in_national_area : national economic area
    // in_industry_count : industry-total establishment count
    // in_dist : interregional distance data
    
    ~ClusterBuilder();
};




#endif
