//
//  ClusterSystem.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#include "ClusterSystem.h"


ClusterSystem::
ClusterSystem( double in_total_area,
              double in_total_count,
              set<Region*>& in_residual_regions )
{
    total_area = in_total_area;
    total_count = in_total_count;
    Cluster::total_area = total_area;
    Cluster::total_count = total_count;
    
    for ( set<Region*>::iterator i = in_residual_regions.begin();
         i != in_residual_regions.end(); ++i )
    {
        residual_regions.insert( *i );
    }
}

void ClusterSystem::
AddMember( Cluster* new_cluster )
{
    members.insert( new_cluster );
    ShrinkResidualRegions( new_cluster );
    Update();
}

bool ClusterSystem::
operator==(  ClusterSystem& in_system )
{
    
    if ( MembersC().size() != in_system.MembersC().size() )
        return false;
    
    if ( residual_regions.size() != in_system.ResidualRegions().size() )
        return false;
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        bool if_found = false;
        
        for ( set<Cluster*>::iterator i2 = 
             in_system.MembersC().begin();
             i2 != in_system.MembersC().end(); ++i2 )
        {
            if ( **i1 == **i2 )
                if_found = true;
        }
        
        if (  !if_found )
            return false;
    }
    
    return true;
}

ClusterSystem& ClusterSystem::
operator=( ClusterSystem& in_system )
{
    total_area = in_system.TotalArea();
    total_count = in_system.TotalCount();
    
    members.clear();
    
    for ( set<Cluster*>::iterator i = in_system.MembersC().begin();
         i != in_system.MembersC().end(); ++i )
    {
        Cluster* new_cluster = new Cluster( **i );
        members.insert( new_cluster );
    }
    
    criterion = in_system.Criterion();
    tc = in_system.Tc();
    
    residual_regions.clear();
    
    
    for ( set<Region*>::iterator i = in_system.ResidualRegions().begin();
         i != in_system.ResidualRegions().end(); ++i )
    {
        residual_regions.insert( *i );
    }
    
    return *this;
}


double ClusterSystem::
CalcTc()
{
    Update();
    double l_hat_c = 0.0;
    double l_0 = 0.0;
    double area_sum = 0.0;
    double count_sum = 0.0;
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        
        l_hat_c += (**i1).Count() * 
        log( (**i1).Count() / total_count );
        
        l_0 += (**i1).Count() * 
        log( (**i1).Area() / total_area );
        
        area_sum += (**i1).Area();
        count_sum += (**i1).Count();
    }
    
    double residual_count = total_count - count_sum;
    double residual_area = total_area - area_sum;
    
    if ( residual_count > 0.0 )
    {
        l_hat_c += residual_count * log ( residual_count / total_count );
        l_0 += residual_count * log ( residual_area / total_area );
    }
    
//    tc = -2.0 * ( l_0 - l_hat_c );
    tc = l_hat_c - l_0;
    
    return tc;
}

double ClusterSystem::
Evaluate()// int in_criterion )
{
    criterion = tc - 0.5 * log( total_count ) * members.size();
    return criterion;
}


//double ClusterSystem::
//Evaluate( map<int, vector<double>* >& penalty_table  )
//{
//    criterion = tc - (*(penalty_table[total_count]))[members.size()-1];
//    
//    return criterion;
//}
//
//void ClusterSystem::
//CalcTcShare()
//{
//    for ( set<Cluster*>::iterator i1 = members.begin();
//         i1 != members.end(); ++i1 )
//    {
//        double l_hat_c = 
//        (**i1).Count() * log( (**i1).Count() / total_count );      
//        
//        double l_0 =
//        (**i1).Count() * log( (**i1).Area() / total_area );
//        
//        (**i1).TcShare( -2.0 * ( l_0 - l_hat_c ) / tc );
//    }
//}

// double ClusterSystem::
// CalcPvalue()
// {
//   return NR::gammp(0.5 * members.size(), 0.5*tc );
// }

// double ClusterSystem::
// CalcPvalueRelative( double ref_tc )
// {
//   if ( tc == ref_tc )
//     return 0.0;
//   else
//     return NR::gammp( 0.5 * members.size() - 1 , 0.5 * ( tc - ref_tc ) );
// }

void ClusterSystem::
Erase( Cluster* target )
{
    members.erase( target );
}

void ClusterSystem::
EraseMember( set<Cluster*>& delete_list )
{
    for ( set<Cluster*>::iterator i1 = delete_list.begin();
         i1 != delete_list.end(); ++i1 )
    {
        //      ExpandResidualRegions( *i1 );
        members.erase(*i1);
        delete *i1;
    }
    
    Update();
}

void ClusterSystem::
EraseMember( Cluster* target )
{
    members.erase( target );
    delete target;
}

void ClusterSystem::
Update()
{
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        (**i1).CalcCountAndArea();
    }
}

void ClusterSystem::
ShrinkResidualRegions( Cluster* new_cluster )
{
    for ( set<Region*>::iterator i = new_cluster->DConvexSolid().begin();
         i != new_cluster->DConvexSolid().end(); ++i )
    {
        residual_regions.erase( *i );
    }
}

//void ClusterSystem::
//ExpandResidualRegions( Cluster* delete_cluster )
//{
//    for ( set<Region*>::iterator i1 =
//         delete_cluster->DConvexSolid().begin(); 
//         i1 != delete_cluster->DConvexSolid().end(); ++i1 )
//    {
//        residual_regions.insert( *i1 );
//        
//        for ( set<Region*>::iterator i2 =
//             (**i1).Neighbors().begin();
//             i2 != (**i1).Neighbors().end(); ++i2 )
//        {
//            
//            for ( set<Cluster*>::iterator i3 =
//                 members.begin(); 
//                 i3 != members.end(); ++i3 )
//            {
//                if ( delete_cluster->Location() == (**i3).Location() )
//                    continue;
//                
//                if ( !((**i3).IfAdjacent( *i2 )) )
//                    residual_regions.insert( *i2 );
//            }
//            
//        }
//    }
//}


double ClusterSystem::
BIC_Contribution(Cluster* adding_cluster,set<Region*>& original_residual_regions, double& original_residual_bic )
{
    set<Region*> tmp_residual_regions;
        
    for ( set<Region*>::iterator i = original_residual_regions.begin();
         i != original_residual_regions.end(); ++i )
    {
        if ( !(adding_cluster->IfDConvexSolidMember( *i ) ) )
            tmp_residual_regions.insert( *i );
    }
    
    double tmp_residual_count = 0.0;
    double tmp_residual_area = 0.0;
    double tmp_residual_bic = 0.0;
    
    for ( set<Region*>::iterator i = tmp_residual_regions.begin();
         i != tmp_residual_regions.end(); ++i )
    {
        tmp_residual_count += (**i).Count();
        tmp_residual_area += (**i).Area();
    }
    
    
    
    if ( tmp_residual_count > 0 )
    {
        tmp_residual_bic = tmp_residual_count * log(tmp_residual_count * total_area / total_count / tmp_residual_area );
    }
    else
    {
        tmp_residual_bic = 0.0;
    }
    
    double cluster_likelihood = adding_cluster->Count() * log( adding_cluster->Count() * total_area / total_count / adding_cluster->Area() );
    
    return (tmp_residual_bic - original_residual_bic + cluster_likelihood - 0.5*log(total_count));

}


double ClusterSystem::
Residual_BIC(set<Region*>& tmp_residuals )
{
    double tmp_count = 0.0;
    double tmp_area = 0.0;
    
    for ( set<Region*>::iterator i1 = tmp_residuals.begin();
         i1 != tmp_residuals.end(); ++i1 )
    {
        tmp_count += (**i1).Count();
        tmp_area += (**i1).Area();
    }
    
    return ( tmp_count * log( tmp_count * total_area /total_count / tmp_area) );
}




void ClusterSystem::
WriteSummary( ofstream& os, set<Region*>& all_regions )
{
    string tmp_str;
    
    os << Conversion( tmp_str, Conversion(tmp_str, "ORDER",0 )+",");
    os << Conversion(tmp_str, Conversion(tmp_str, "CENTER",0)+ ",");
    os << Conversion(tmp_str, Conversion(tmp_str, "COUNT", 0)+ ",");
    os << Conversion(tmp_str, Conversion(tmp_str, "COUNT_SHARE",0)+ ",");
    os << Conversion(tmp_str, Conversion(tmp_str, "AREA_SHARE", 0) + ",");
    os << Conversion(tmp_str, Conversion(tmp_str, "BIC_INC", 0) + ",");
    os << Conversion(tmp_str, "INC_SHARE" ) << endl;
    
    if ( members.size() == 0 )
    {
        os.close();
        os.clear();
        return;
    }
    
    
    list<Cluster*> cluster_list;
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        cluster_list.push_back( *i1 );
    }

    set<Region*> current_residuals;
    for (set<Region*>::iterator i1 = all_regions.begin();
         i1 != all_regions.end(); ++i1 )
    {
        current_residuals.insert(*i1);
    }
    
    double bic_0 = 0.0;
    int included_clusters = 0;
    
    while ( included_clusters < members.size() )
    {
        Cluster* sig_cluster;
        
        if ( cluster_list.size() == 1 )
        {
            sig_cluster = *(cluster_list.begin());
            cluster_list.pop_front();
            sig_cluster->BIC( BIC_Contribution(sig_cluster, current_residuals, bic_0 ) );
            break;
        }
        else
        {
            double max_increment = -DBL_MAX;

            for ( list<Cluster*>::iterator i1 = cluster_list.begin();
                 i1 != cluster_list.end(); ++i1 )
            {
                double val = BIC_Contribution(*i1, current_residuals, bic_0 );
                
                if ( val > max_increment )
                {
                    max_increment = val;
                    sig_cluster = *i1;
                }
            }

            sig_cluster->BIC(max_increment);

            for (set<Region*>::iterator i1 = sig_cluster->DConvexSolid().begin();
                 i1 != sig_cluster->DConvexSolid().end(); ++i1 )
            {
                current_residuals.erase(*i1);
            }
            bic_0 = Residual_BIC(current_residuals );
            ++included_clusters;
            cluster_list.remove(sig_cluster);
        }
    }
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        cluster_list.push_back( *i1 );
    }
    
    cluster_list.sort( ClusterCompareBIC() );
    
    int order_val = 0;
    
    for ( list<Cluster*>::iterator 
         i1 = cluster_list.begin();
         i1 != cluster_list.end(); ++i1 )
    {
        ++order_val;
        
        os << Conversion( tmp_str, Conversion(tmp_str, order_val, 0) + ",");
        os << Conversion( tmp_str, Conversion(tmp_str, (**i1).Location(),0) + ",");
        os << Conversion( tmp_str, Conversion(tmp_str, (**i1).Count(), 0) + ",");
        os << Conversion( tmp_str, Conversion(tmp_str, ( (**i1).Count() / total_count ), 0) + ",");
        os << Conversion( tmp_str, Conversion(tmp_str, ( (**i1).Area() / total_area ), 0) + ",");
        os << Conversion( tmp_str, Conversion(tmp_str, (**i1).BIC(), 0) + ",");
        os << Conversion( tmp_str, ( (**i1).BIC() / criterion ) ) << endl;
    }
    os.close();
    os.clear();
}

void ClusterSystem::
WriteMapData( ofstream& os )
{
    string tmp_str;
    
    os << "CENTER, MEMBER" << endl;
    
    if ( members.size() == 0 )
    {
        os.close();
        os.clear();
        return;
    }
    
    list<Cluster*> cluster_list;
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        cluster_list.push_back( *i1 );
    }
    
    cluster_list.sort( ClusterCompareLocation() );
    
    for ( list<Cluster*>::iterator i1 = cluster_list.begin();
         i1 != cluster_list.end(); ++i1 )
    {
        list<Region*> mbc;
        
        for ( set<Region*>::iterator i2 = 
             (**i1).DConvexSolid().begin();
             i2 != (**i1).DConvexSolid().end(); ++i2 )
        {
            mbc.push_back( *i2 );
        }
        
        mbc.sort( LocationIdCompare() );
        
        for ( list<Region*>::iterator i2 = mbc.begin();
             i2 != mbc.end(); ++i2 )
        {
            os << (**i1).Location() << ", " << (**i2).Location() << endl;
        }
    }
    os.close();
    os.clear();
}

void ClusterSystem::
WriteCore( ofstream& os )
{
    string tmp_str;
    
    os << "CENTER, CORE" << endl;
    
    if ( members.size() == 0 )
    {
        os.close();
        os.clear();
        return;
    }
    
    list<Cluster*> cluster_list;
    
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        cluster_list.push_back( *i1 );
    }
    
    cluster_list.sort( ClusterCompareLocation() );
    
    for ( list<Cluster*>::iterator i1 = cluster_list.begin();
         i1 != cluster_list.end(); ++i1 )
    {
        
        list<Region*> core_out;
        
        for ( set<Region*>::iterator i2 = 
             (**i1).Core().begin();
             i2 != (**i1).Core().end(); ++i2 )
        {
            core_out.push_back( *i2 );
        }
        
        core_out.sort( LocationIdCompare() );
        
        for ( list<Region*>::iterator i2 = core_out.begin();
             i2 != core_out.end(); ++i2 )
        {
            os << (**i1).Location() << ", " << (**i2).Location() << endl;
        }
    }
    os.close();
    os.clear();
}


ClusterSystem::
~ClusterSystem()
{
    for ( set<Cluster*>::iterator i1 = members.begin();
         i1 != members.end(); ++i1 )
    {
        if ( *i1 != NULL )
            delete *i1;
    }
}

