//
//  ClusterSystem.h
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_ClusterSystem_h
#define DetectClusters_ClusterSystem_h

#include <set>
#include <list>
#include <algorithm>
#include <cmath>
#include <iostream>
#include <fstream>
#include "Region.h"
#include "Cluster.h"
#include "Distance.h"
#include "StringManipulator.h"

using namespace std;

class ClusterSystem    
{
private:
    double total_area;
    // National economic area.
    
    double total_count;
    // Total establishment count in the industry.
    
    set<Cluster*> members;
    // Set of all clusters.
    
    set<Region*> residual_regions;
    // Set of residual regions.
    
    double criterion;
    // Criterion value (BIC or AIC)
    
    double tc;
    // Model likelihood.
    
public:
//    enum criterion { _aic,
//        _bic,
//   //     _nml,
//        _criterion_end };
//    
    
public:
    
    double TotalArea() const { return total_area; }
    double TotalCount() const { return total_count; }
    
    set<Cluster*>& Members() { return members; }
    const set<Cluster*>& MembersC() const { return members; }
    
    set<Region*>& ResidualRegions() { return residual_regions; }
    
    void AddMember( Cluster* new_cluster );
    // Add new_cluster to members. 
    // Call Update()
    
    void Erase( Cluster* target );
    // Erase target from member cluster.
    
    void EraseMember( set<Cluster*>& delete_list );
    // Deletes clusters in delete_list from cluster scheme (members).

    void EraseMember( Cluster* target );
    // Deletes target from cluster scheme (members).
    
    double CalcTc();
    // Computes the model likelihood.
    
    double Tc() const { return tc; }
    void Tc( double in_val ) { tc = in_val; }
    
    void Update() ;
    // Updates establishment count (count) and economic area (area) of each member.
    
    void ShrinkResidualRegions( Cluster* new_cluster );
    // Removes basic regions of new cluster from residual_regions.
    
//    void ExpandResidualRegions( Cluster* delete_cluster );
    // Returns d_convex_solid in delete_cluster to residual regions.
    // Expand residual_regions to include delete_cluster and its 
    // neighboring regions on condition that these are non-adjacent
    // to the remaining clusters.
    
    
    double Evaluate();
    // Compute BIC value, and save it to criterion.
    
//    double Evaluate( map<int, vector<double>* >& penalty_table );
    // NML version of Evaluate.
    
    
    double Criterion() const { return criterion; }
    
    ClusterSystem& operator=( ClusterSystem& in_system );
    // Copies:
    //  members
    //  residual_regions
    //  tc
    
    bool operator==( ClusterSystem& in_system );
    // Returns true if members are the same.
    
    double BIC_Contribution(Cluster* adding_cluster,
                     set<Region*>& original_residual_regions, 
                     double& original_residual_bic );
    // Computes BIC contribution of a given cluster (adding_cluster)
    // in the cluster scheme.
    // original_residual_regions : residual regions before adding adding_cluster.
    // original_residual_bic : BIC value before adding adding_cluster.

    double Residual_BIC(set<Region*>& tmp_residuals );
    // BIC contribution of residual regions (tmp_residuals).
    
    ///// I/O /////
    
    void WriteSummary( ofstream& os, set<Region*>& all_regions );
    // List of clusters with summary information:
    // 1) Order in terms of BIC contribution 
    // 2) Center location 
    // 3) Establishment count
    // 4) Establishment count share (relative to industry total)
    // 5) Economic area share (relative to national total)
    // 6) BIC contribution
    // 7) BIC contribution share
    
    void WriteMapData( ofstream& os );
    // Center location, d-convex solid members
    
    void WriteCore( ofstream& os );
    // Center location, core members
    
    ClusterSystem( double in_total_area,
                  double in_total_count,
                  set<Region*>& in_residual_regions );
    
    ClusterSystem( ClusterSystem& in_system )
    { operator=( in_system ); }
    
    ~ClusterSystem();
};



#endif
