//
//  Region.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//


#include "Region.h"

Region::
Region( int loc,
       double in_area)
{
    location = loc;
    area = in_area;
    count = 0.0;
    if_border = false;
}

Region::
Region( const Region& in_region )
{
    operator=(in_region);
}

Region& Region::
operator=( const Region& in_region )
{
    location = in_region.Location();
    area = in_region.Area();
    area_share = in_region.AreaShare();
    count = in_region.Count();
//    share = in_region.Share();
    if_border = in_region.IfBorder();
    return *this;
}

void Region::
AddNeighbor( Region* in_region )
{
    neighbors.insert( in_region );
}


//void Region::
//Share( double in_total )
//{ 
//    share = count / in_total; z
//}


int Region::
FindShortestPath( set<Region*>& path,
                 Region* target )
{
    if ( *target == *this )
        return 1;
    if ( shortest_path[target] == NULL )
        return 0;
    
    Region*& next_move = FindLocalMove( target );
    path.insert( next_move );
    return next_move->FindShortestPath( path, target );
}


bool Region::
operator==( const Region& in_region ) const
{
    return in_region.Location() == Location();
}

bool Region::
operator==( const Region* in_region ) const
{
    return in_region->Location() == Location();
}

bool Region::
IfOutSideConvexHull( set<Region*>& region_set, set<Region*>& path )
{
    // If this function is called, this region is supposed to be outside the region_set.

    if ( if_border )
    {
        for ( set<Region*>::iterator i1 = neighbors.begin(); 
             i1 != neighbors.end(); ++i1 )
        {
            if ( region_set.find(*i1) == region_set.end() )
            {
                // At least one of the neighbors is outside region_set.
                return true;
            }
            
            // If this border region is completely surrounded by
            // region_set then the fist region which called IfOutSideConvexHull()
            // is considered to be interior to the region_set.
            // But, if there is at least one neighbor of this border region
            // which is not in the region_set, it is considered to be exterior to the region_set.
        }      
        return false;
    }
    
    // If this is not at the border, then move to neighbor region
    // which is not included in region_set nor the path of regions
    // from the first region which called IfOutSideConvexHull().
    
    path.insert( this );
    // A sequence of adjacent regions up to this region outside region_set.
    
    bool if_outside = false;
    
    for ( set<Region*>::iterator i1 = neighbors.begin(); 
         i1 != neighbors.end(); ++i1 )
    {
        if ( region_set.find(*i1) != region_set.end() )
            continue;
        
        if ( path.find( *i1 ) != path.end() )
            continue;
        
        if_outside = (**i1).IfOutSideConvexHull( region_set, path );
        // This neighbor region is neither in region_set nor in the path.
        
        if ( if_outside )
            // If some path from this region reaches the border, and if
            // this region is not completely surrounded by region_set, then
            // this region is outside region_set.
            return true;
        
    }
    
    // There is no way out to get to the border
    return false;
}


void Region::
Read( ShortestPathData& in_data,
     map<int, Region*>& region_id )
{
    for ( map<int, int>::iterator i = in_data(location).begin();
             i != in_data(location).end(); ++i )
    {
        if (i->second < 0)
            shortest_path[region_id[i->first]] = NULL;
        else
            shortest_path[ region_id[i->first] ] = region_id[i->second];
    }
}


bool LocationIdCompare::
operator()( const Region* r1, const Region* r2 ) const
{
    return r1->Location() < r2->Location();
}

bool DistanceCompareRegion::
operator() ( const Region* r1, const Region* r2 ) const
{
    return ( dist( r1->Location(), ref_region->Location() ) < 
            dist( r2->Location(), ref_region->Location() ) );
}

bool ValueCompareRegion::
operator() ( pair<Region*, double> r1,
            pair<Region*, double> r2 ) const
{
    return ( r1.second < r2.second );
}
