//
//  Region.h
//  
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_Region_h
#define DetectClusters_Region_h

#include <set>
#include <vector>
#include <cmath>
#include <iostream>
#include "Distance.h"
#include "ShortestPathData.h"

using namespace std;

class Region
{
private:
    
    int location;
    // Location ID.
    
    double count;
    // Number of establishments in this region
    
    double area;
    // Areal size
    
    double area_share;
    // Share of this region in the total national area.
        
    bool if_border;
    // True if this region is at the border
    
    set<Region*> neighbors;
    // Set of adjacent regions
    
    map<Region*, Region*> shortest_path;
    // shortest_path[destination region] = 1st region on the shortest path to 
    // the destination region.
    
public:
    int Location() const { return location; }
    
    double Area() const { return area; }
    void Area( double val ) { area = val; }
    void SetAreaShare( double in_total ) { area_share = area / in_total; }
    double& SetAreaShare() { return area_share; }
    double AreaShare() const { return area_share; }
    
    double Count() const { return count; }
    void Count( double in_val ) { count = in_val; }
    double& SetCount() { return count; }
    
    bool IfBorder() const { return if_border; }
    void SetBorder() { if_border = true; }
    
    const set<Region*>& Neighbors() const { return neighbors; }
    void AddNeighbor( Region* in_region );
    set<Region*>& SetNeighbors() { return neighbors; }
    
    ///  double CalcPvalue( double total_count );
    
    Region*& FindLocalMove( Region* target )
    { return shortest_path[target]; }
    // Returns the first region on the shortest path to target region.
    
    
    int FindShortestPath( set<Region*>& path,
                          Region* target );
    
    // Find the shortest region-path from this to target.
    

    bool IfOutSideConvexHull( set<Region*>& region_set, set<Region*>& path );
    // Returns true if this region is exterior to region_set.
    // Condition:
    //  i) There is a sequence of adjacent regions which are exterior to region_set from this region to the border.
    //  ii) If this region is at the border and 
    // Find if there is a path from this region to the border outside region_set.
    //
    // Used to solidify d-convex set of regions.
    
    void Read( ShortestPathData& in_data,
              map<int, Region*>& region_id );	
    // Set shortest_path.
    
    bool operator==( const Region& in_region ) const;
    bool operator==( const Region* in_region ) const;
    // True if location is the same.
    
    Region& operator=( const Region& in_region );
    // Copy location and area of in_region.
    
    Region( int loc, double in_area = 0.0 ) ;
    // 1st input = location ID
    // 2nd input = areal size
    
    Region( const Region& in_region );
    // Construct a region with location and area of in_region.
    // (Call operator=().)
    
    ~Region() {}
    
};


class LocationIdCompare
{
public:
    bool operator()( const Region* r1, const Region* r2 ) const;
	
};

class DistanceCompareRegion
{
public:
    Region* ref_region;
    Distance& dist;
    DistanceCompareRegion(  Distance& in_dist, Region* in_ref ) :
 	dist( in_dist ), ref_region( in_ref ) {}
    ~DistanceCompareRegion() {}
    
    bool operator()( const Region* r1, const Region* r2 ) const;
};    



class ValueCompareRegion
{
public:
    ValueCompareRegion() {}
    ~ValueCompareRegion() {}
    
    bool operator() ( pair<Region*, double> r1,
                     pair<Region*, double> r2 ) const;
};




#endif
