#
# ShortestPathLocalMoveMP.py
#
# Derive shortest path between each pair of connected regions, and
# the corresponding shortest-path distance.
#
# Output (1) : Shortest path between each pair of connected regions
# 
#   0th column = List of all regions.
#   Element (i,j) : the next region to visit from region i
#                   on the shortest path from region i to region (j-1)
#                   (If the two regions are disconnected, this entry is "-1".)
#
# Output (2) : Shortest-path distance between each pair of connected regions.
#   1st column = Origin region
#   2nd column = Destination region
#   3rd column = Shortest-path distance
#   (Distance is computed only between connected pairs, i.e., if there is no
#   entry for a given pair of regions, these regions are not connected;
#   Within-region distance is defined to be zero.)
#

country = 'de'
year = '2008'
num_processes = 12
filetail = '_' + country + '_' + year + '.csv'

### INPUTS ###

basedir = '../DATA/' + country + '/basic_regions/'
inname = basedir + 'neighbors' + filetail
# Two column data : (Region, its neighbor)

region_file = basedir + 'unions_economic_area' + filetail
# A column of all regions (No title row)

dist_file = basedir + 'travel_minutes_between_unions' + filetail
# Three columns data: (Origin, Destination, Distance) with a title row

out_path = basedir + 'shortest_path_local_move' + filetail
# Output (1)

out_dist = basedir + 'shortest_path_distance' + filetail
# Output (2)



import csv, sys
import multiprocessing
from multiprocessing import Pool, Lock, Value
from datetime import datetime


task_start = datetime.now()

### Read regions ###

infile = open(region_file,'r')
reader = csv.reader(infile)

regions = []
for row in reader:
    regions.append(int(row[0]))
    
infile.close

regions.sort()




### Read neighbors ###

infile = open(inname,'r')
reader = csv.reader(infile)

reader.next()
neighbors = {}

for row in reader:
    [center,neighbor] = map(int,row)
    
    ## TEST
    if not center in regions:
        continue
    
   
    if (center in neighbors) == False:
        neighbors[center] = set()
        
    #TEST        
    if not neighbor in regions:
        continue
 
    neighbors[center].add(neighbor)

infile.close



## Connected sets ###

connected_regions = {}
    
def find_connected_regions(x, result):
    result.add(x)
    num_added = 1
    
    while (num_added > 0):
        expansion = set()
        
        for r in result:
            if r in neighbors:
                expansion |= neighbors[r]
        
        expansion -= result
        num_added = len(expansion)
        result |= expansion
        

for r in regions:
    connected_regions[r] = set()
    
    if (r in neighbors) == False:
        connected_regions[r].add(r)
        # Region r is isolated.
        continue
    else:
        find_connected_regions(r,connected_regions[r])
    

connected_num = {}

for r in regions:
    value = len(connected_regions[r])
    if (value in connected_num) == False:
        connected_num[value] = []
        
    connected_num[value].append(r)

connected_nums = connected_num.keys()
connected_nums.sort()
connected_nums.reverse()
print '\nNumber of connected regions ...'

for value in connected_nums:
    print value, ' : '

    connected_num[value].sort()

    for r in connected_num[value]:
        print r,
    
    print ''
    print ''


### Distances ###

infile = open(dist_file,'r')
reader = csv.reader(infile)
reader.next()

dist_values = {}

for row in reader:
    [orig,dest] = map(int,row[:2])

    ## TEST
    if not orig in connected_regions:
        continue
        
    if (orig in dist_values) == False:
        dist_values[orig] = {}
        dist_values[orig][orig] = 0.0

    if (dest in dist_values) == False:
        dist_values[dest] = {}
        dist_values[dest][dest] = 0.0
         
    if (dest in connected_regions[orig]) == False:
        continue
        # Distance between non-connected regions are reset
        # to infinity (even if a finite value is provided by the data)
    
   
    dist_values[orig][dest] = float(row[2])

infile.close


for i in range(len(regions)):
    for j in range(i+1, len(regions)):
        if (regions[j] in dist_values[regions[i]]) == False:
            dist_values[regions[i]][regions[j]] = float('inf')
            
# If there is no data on distance between a given region pair, then
# these regions are considered to be disconnected.
   
def distance(x,y):
    if x <= y:
        return dist_values[x][y]
    else:
        return dist_values[y][x]


### Initialization ###

shortest_path = {}
shortest_distance = {}

for r in regions:
    shortest_path[r] = {}
    shortest_distance[r] = {}

    shortest_distance[r][r] = 0.0
    shortest_path[r][r] = r
    
    for s in regions:
        if r == s:
            continue
        
        shortest_path[r][s] = -1
        shortest_distance[r][s] = float('inf')




### Shortest path ###
#completed_count = 0


#def CompState(l,arg):
#    global completed_count
#    l.acquire()
#    completed_count+=1
#    report = [completed_count]
#    report.extend(arg)
#    print report
#    l.release()
#    
class Counter(object):
    def __init__(self, initval=0):
        self.val = Value('i', initval)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value
        
        
def IncCount(arg):
    global completed_count
    global lock
    
    lock.acquire()
    completed_count += 1
    report = str(completed_count) + arg
    print report
    lock.release()
  
    
def find_shortest_path(r1):
    global lock, completed_count
    local_task_start = datetime.now()
    result_path = []
    result_dist = []
    
    determined = []
    determined.append(r1)
    
    shortest_distance_local = {}
    for r in regions:
        if not r == r1:
            shortest_distance_local[r] = float('inf')
        else:
            shortest_distance_local[r] = 0.0
    
    if len(connected_regions[r1]) == 1:
        result_dist.append([r1,0.0])
        result_path.append([r1,r1])
    else:
        
        while ( len(determined) < len(connected_regions[r1])):
            min_dist = float('inf')
            base = r1
            move = r1
            
            for r2 in determined:
                for r3 in neighbors[r2]:
                    if r3 in determined:
                        continue
                    
                    tmp_dist = distance(r2,r3) + shortest_distance_local[r2]
                    
                    if tmp_dist < min_dist:
                        min_dist = tmp_dist
                        base = r2
                        move = r3
                        
    #        shortest_distance[r1][move] = min_dist
    #        shortest_path[r1][move] = base
    
    #        if base == move:
    #            min_dist = 0.0
            shortest_distance_local[move] = min_dist
            result_dist.append( [move, min_dist])
            result_path.append( [move, base])
            
            
            #if r1 == 1058135:
            #    print move, min_dist, base
            #    
            determined.append(move)
            
        
    local_task_end = datetime.now()
    completed_count.increment()
#    print outstr = '(' + str(local_task_end - task_start) + ') : Region ' + str(r1) +  '(' + str(local_task_end - local_task_start) + ')'
#    if completed_count.value() % 100 == 0:
    print completed_count.value(),'(', local_task_end - task_start, ') : Region ' , r1, '(', local_task_end - local_task_start, ')'

#    IncCount(outstr)
    
    return [result_dist, result_path]


def find_all_shortest_paths():
    pool = Pool(num_processes)
    return pool.map(find_shortest_path, regions)
    
if not __name__ == '__main__':
    sys.exit()
    
completed_count = Counter(0)
lock = Lock()


result = find_all_shortest_paths()

measure_stop = datetime.now()

for i in range(len(regions)):
    for j in range(len(result[i][0])):
        shortest_distance[regions[i]][result[i][0][j][0]] = result[i][0][j][1]
        shortest_path[regions[i]][result[i][1][j][0]] = result[i][1][j][1]

#for r1 in regions:
#    
#    sys.stdout.write(str(r1) + ' ')
#    sys.stdout.flush()
#    
    
    
#### From here ###    
    #determined = []
    #determined.append(r1)
    #
    #while ( len(determined) < len(connected_regions[r1])):
    #    min_dist = float('inf')
    #    base = r1
    #    move = r1
    #    
    #    for r2 in determined:
    #        for r3 in neighbors[r2]:
    #            if r3 in determined:
    #                continue
    #            
    #            tmp_dist = distance(r2,r3) + shortest_distance[r1][r2]
    #            
    #            if tmp_dist < min_dist:
    #                min_dist = tmp_dist
    #                base = r2
    #                move = r3
    #                
    #                
    #    shortest_distance[r1][move] = min_dist
    #    shortest_path[r1][move] = base
    #    determined.append(move)
    #    
    ### To here ###

outfile_dist = open(out_dist,'w')
writer_dist = csv.writer(outfile_dist)
writer_dist.writerow(['ORIG','DEST','DIST'])

for r1 in regions:

    connected_regions[r1] = list(connected_regions[r1])
    connected_regions[r1].sort()
    
    for r2 in connected_regions[r1]:
        if r1 > r2:# or (r2 in shortest_distance[r1]) == False:
            continue
        
        data = [r1]
        
        writer_dist.writerow([r1,r2,shortest_distance[r1][r2]])
    

outfile_dist.close


outfile_path = open(out_path, 'w')
writer_path = csv.writer(outfile_path)
header = ['REGION']
header.extend(regions)
writer_path.writerow(header)

for r1 in regions:

    data = [r1]

    for r2 in regions:

        
        if r2 in connected_regions[r1]:
            data.append(shortest_path[r2][r1])
        else:
            data.append(-1)
            
    writer_path.writerow(data)



outfile_path.close


print "Elapsed time = ", datetime.now() - task_start 


        
        
        
        
    





